import json
import os
import random
import re
import string
import sys
from typing import Any, Callable, Dict, List, Optional
from unittest.mock import Mock, patch

import boto3
from botocore.exceptions import ClientError
from click import ClickException
from moto import mock_cloudformation, mock_ec2, mock_efs, mock_iam, mock_s3
import pytest

from anyscale.aws_iam_policies import (
    AMAZON_ECR_READONLY_ACCESS_POLICY_ARN,
    AMAZON_S3_FULL_ACCESS_POLICY_ARN,
    ANYSCALE_IAM_PERMISSIONS_EC2_INITIAL_RUN,
    ANYSCALE_IAM_PERMISSIONS_EC2_STEADY_STATE,
    ANYSCALE_IAM_PERMISSIONS_SERVICE_STEADY_STATE,
    get_anyscale_aws_iam_assume_role_policy,
)
from anyscale.cli_logger import CloudSetupLogger
from anyscale.client.openapi_client.models.aws_memory_db_cluster_config import (
    AWSMemoryDBClusterConfig,
)
from anyscale.client.openapi_client.models.create_cloud_resource import (
    CreateCloudResource,
)
from anyscale.client.openapi_client.models.subnet_id_with_availability_zone_aws import (
    SubnetIdWithAvailabilityZoneAWS,
)
from anyscale.cloud_resource import (
    verify_aws_cloudformation_stack,
    verify_aws_efs,
    verify_aws_iam_roles,
    verify_aws_memorydb_cluster,
    verify_aws_s3,
    verify_aws_security_groups,
    verify_aws_subnets,
    verify_aws_vpc,
)
from anyscale.conf import ANYSCALE_IAM_ROLE_NAME
from anyscale.shared_anyscale_utils.aws import AwsRoleArn
from anyscale.shared_anyscale_utils.conf import ANYSCALE_HOST
from anyscale.utils.network_verification import AWS_SUBNET_CAPACITY


DEFAULT_RAY_IAM_ROLE = "ray-autoscaler-v1"


def _create_roles(anyscale_account: str):
    iam = boto3.client("iam")
    suffix = "".join(random.choices(string.ascii_lowercase, k=10))
    control_plane_role = iam.create_role(
        RoleName=ANYSCALE_IAM_ROLE_NAME + suffix,
        AssumeRolePolicyDocument=json.dumps(
            get_anyscale_aws_iam_assume_role_policy(
                anyscale_aws_account=anyscale_account
            )
        ),
    )["Role"]["Arn"]

    # Create & Configure the Dataplane Role
    dataplane_role_name = DEFAULT_RAY_IAM_ROLE + suffix
    dataplane_role = iam.create_role(
        RoleName=dataplane_role_name,
        AssumeRolePolicyDocument=json.dumps(
            {
                "Version": "2012-10-17",
                "Statement": [
                    {
                        "Sid": "Allow",
                        "Effect": "Allow",
                        "Principal": {"Service": "ec2.amazonaws.com"},
                        "Action": "sts:AssumeRole",
                    }
                ],
            }
        ),
    )["Role"]["Arn"]

    iam.attach_role_policy(
        RoleName=dataplane_role_name, PolicyArn=AMAZON_ECR_READONLY_ACCESS_POLICY_ARN
    )
    iam.attach_role_policy(
        RoleName=dataplane_role_name, PolicyArn=AMAZON_S3_FULL_ACCESS_POLICY_ARN
    )
    return [control_plane_role, dataplane_role]


def _delete_role(role_arn: str) -> None:
    iam = boto3.client("iam")
    role_name = AwsRoleArn.from_string(role_arn).to_role_name()

    policies = iam.list_attached_role_policies(RoleName=role_name)["AttachedPolicies"]

    for policy in policies:
        iam.detach_role_policy(RoleName=role_name, PolicyArn=policy["PolicyArn"])

    iam.delete_role(RoleName=role_name)


def generate_cloud_resource_mock_aws(
    enable_head_node_fault_tolerance: bool = False,
) -> CreateCloudResource:
    return CreateCloudResource(
        aws_vpc_id="fake_aws_vpc_id",
        aws_subnet_ids_with_availability_zones=[
            SubnetIdWithAvailabilityZoneAWS(
                subnet_id="fake_aws_subnet_id_0", availability_zone="fake_aws_az_0"
            ),
            SubnetIdWithAvailabilityZoneAWS(
                subnet_id="fake_aws_subnet_id_1", availability_zone="fake_aws_az_1"
            ),
        ],
        aws_iam_role_arns=[
            "arn:aws:iam::123:role/mock_anyscale_role",
            "arn:aws:iam::123:role/mock_dataplane_role",
        ],
        aws_security_groups=["fake_aws_security_group_0"],
        aws_s3_id="fake_aws_s3_id",
        aws_efs_id="fake_aws_efs_id",
        aws_cloudformation_stack_id="fake_aws_cloudformation_stack_id",
        memorydb_cluster_config=AWSMemoryDBClusterConfig(
            id="fake_memorydb_id", endpoint="fake_memorydb_endpoint:6379"
        )
        if enable_head_node_fault_tolerance
        else None,
    )


def _attach_iam_roles_to_role(role_arn: str) -> None:
    iam = boto3.client("iam")
    role_name = AwsRoleArn.from_string(role_arn).to_role_name()

    steady_state = iam.create_policy(
        PolicyName="".join(random.choices(string.ascii_lowercase, k=10)),
        PolicyDocument=json.dumps(ANYSCALE_IAM_PERMISSIONS_EC2_STEADY_STATE),
    )["Policy"]["Arn"]
    iam.attach_role_policy(RoleName=role_name, PolicyArn=steady_state)

    service_steady_state = iam.create_policy(
        PolicyName="".join(random.choices(string.ascii_lowercase, k=10)),
        PolicyDocument=json.dumps(ANYSCALE_IAM_PERMISSIONS_SERVICE_STEADY_STATE),
    )["Policy"]["Arn"]
    iam.attach_role_policy(RoleName=role_name, PolicyArn=service_steady_state)

    initial_run = iam.create_policy(
        PolicyName="".join(random.choices(string.ascii_lowercase, k=10)),
        PolicyDocument=json.dumps(ANYSCALE_IAM_PERMISSIONS_EC2_INITIAL_RUN),
    )["Policy"]["Arn"]
    iam.attach_role_policy(RoleName=role_name, PolicyArn=initial_run)


def _attach_inline_iam_policy_to_role(role_arn: str) -> None:
    iam = boto3.client("iam")
    role_name = AwsRoleArn.from_string(role_arn).to_role_name()

    iam.put_role_policy(
        RoleName=role_name,
        PolicyName="".join(random.choices(string.ascii_lowercase, k=10)),
        PolicyDocument=json.dumps(ANYSCALE_IAM_PERMISSIONS_EC2_STEADY_STATE),
    )
    iam.put_role_policy(
        RoleName=role_name,
        PolicyName="".join(random.choices(string.ascii_lowercase, k=10)),
        PolicyDocument=json.dumps(ANYSCALE_IAM_PERMISSIONS_SERVICE_STEADY_STATE),
    )
    iam.put_role_policy(
        RoleName=role_name,
        PolicyName="".join(random.choices(string.ascii_lowercase, k=10)),
        PolicyDocument=json.dumps(ANYSCALE_IAM_PERMISSIONS_EC2_INITIAL_RUN),
    )


def _create_instance_profile_for_role(
    role_arn: str, instance_profile_name: Optional[str]
) -> None:
    role_name = AwsRoleArn.from_string(role_arn).to_role_name()
    if instance_profile_name is None:
        instance_profile_name = role_name

    iam = boto3.client("iam")
    iam.create_instance_profile(InstanceProfileName=instance_profile_name)
    iam.add_role_to_instance_profile(
        RoleName=role_name, InstanceProfileName=instance_profile_name,
    )


def _create_vpc(cidr_block: str = "10.0.0.0/16", region: str = "us-west-2") -> str:
    ec2 = boto3.client("ec2", region_name=region)
    vpc = ec2.create_vpc(CidrBlock=cidr_block)
    return vpc["Vpc"]["VpcId"]


def _create_subnets(
    vpc_id: str,
    region: str,
    num_subnets: int = 2,
    assign_public_ip: bool = False,
    use_single_availability_zone: bool = False,
) -> List[SubnetIdWithAvailabilityZoneAWS]:
    ec2 = boto3.client("ec2", region_name=region)
    subnet_ids = []
    for i in range(num_subnets):
        subnet = ec2.create_subnet(
            VpcId=vpc_id,
            CidrBlock=f"10.0.{i}.0/24",
            AvailabilityZone=f"{region}a"
            if use_single_availability_zone
            else f"{region}{string.ascii_lowercase[i]}",
        )
        subnet_id = subnet["Subnet"]["SubnetId"]
        subnet_az = subnet["Subnet"]["AvailabilityZone"]
        subnet_ids.append(
            SubnetIdWithAvailabilityZoneAWS(
                subnet_id=subnet_id, availability_zone=subnet_az
            )
        )
        # MapPublicIpOnLaunch is by default False.
        if assign_public_ip:
            ec2.modify_subnet_attribute(
                SubnetId=subnet_id, MapPublicIpOnLaunch={"Value": True},
            )

    return subnet_ids


def _create_security_group(region: str) -> str:
    ec2 = boto3.client("ec2", region_name=region)
    vpc_id = _create_vpc(region=region)

    security_group = ec2.create_security_group(
        VpcId=vpc_id, GroupName="mock_security_group", Description="mock_security_group"
    )
    security_group_id = security_group["GroupId"]
    ec2.authorize_security_group_ingress(
        GroupId=security_group_id,
        IpPermissions=[
            {
                "IpProtocol": "-1",
                "FromPort": -1,
                "ToPort": -1,
                "UserIdGroupPairs": [{"GroupId": security_group_id,}],
            }
        ],
    )

    return security_group_id


def _create_security_group_given_permissions(
    permissions: List[str], region: str
) -> str:
    ec2 = boto3.client("ec2", region_name=region)
    vpc_id = _create_vpc(region=region)

    security_group = ec2.create_security_group(
        VpcId=vpc_id, GroupName="mock_security_group", Description="mock_security_group"
    )
    security_group_id = security_group["GroupId"]

    port_number = {"https": 443, "ssh": 22, "self": -1, "excess": 123}

    ec2.authorize_security_group_ingress(
        GroupId=security_group_id,
        IpPermissions=[
            {
                "IpProtocol": "-1" if permission == "self" else "tcp",
                "FromPort": port_number[permission],
                "ToPort": port_number[permission],
                "UserIdGroupPairs": [{"GroupId": security_group_id}],
            }
            for permission in permissions
        ],
    )

    return security_group_id


def _create_security_group_inbound_permission_expected(region: str) -> str:
    ec2 = boto3.client("ec2", region_name=region)
    vpc_id = _create_vpc(region=region)

    security_group = ec2.create_security_group(
        VpcId=vpc_id, GroupName="mock_security_group", Description="mock_security_group"
    )
    security_group_id = security_group["GroupId"]
    ec2.authorize_security_group_ingress(
        GroupId=security_group_id,
        IpPermissions=[
            {
                "IpProtocol": "tcp",  # use tcp instead of -1 because of moto impl issue
                "FromPort": 443,
                "ToPort": 443,
                "UserIdGroupPairs": [{"GroupId": security_group_id}],
            },
            {
                "IpProtocol": "tcp",  # use tcp instead of -1 because of moto impl issue
                "FromPort": 22,
                "ToPort": 22,
                "UserIdGroupPairs": [{"GroupId": security_group_id}],
            },
            {
                "IpProtocol": "-1",  # -1 is needed for check here: https://github.com/anyscale/product/blob/master/frontend/cli/anyscale/cloud_resource.py#L378
                "FromPort": -1,
                "ToPort": -1,
                "UserIdGroupPairs": [{"GroupId": security_group_id}],
            },
        ],
    )
    return security_group_id


def _create_security_group_inbound_permission_excessive_permission(region: str) -> str:
    ec2 = boto3.client("ec2", region_name=region)
    vpc_id = _create_vpc(region=region)

    security_group = ec2.create_security_group(
        VpcId=vpc_id, GroupName="mock_security_group", Description="mock_security_group"
    )
    security_group_id = security_group["GroupId"]
    ec2.authorize_security_group_ingress(
        GroupId=security_group_id,
        IpPermissions=[
            {
                "IpProtocol": "tcp",  # use tcp instead of -1 because of moto impl issue
                "FromPort": 443,
                "ToPort": 443,
                "UserIdGroupPairs": [{"GroupId": security_group_id}],
            },
            {
                "IpProtocol": "tcp",  # use tcp instead of -1 because of moto impl issue
                "FromPort": 22,
                "ToPort": 22,
                "UserIdGroupPairs": [{"GroupId": security_group_id}],
            },
            {
                "IpProtocol": "tcp",  # use tcp instead of -1 because of moto impl issue
                "FromPort": 123,
                "ToPort": 123,
                "UserIdGroupPairs": [{"GroupId": security_group_id}],
            },
            {
                "IpProtocol": "-1",  # -1 is needed for check here: https://github.com/anyscale/product/blob/master/frontend/cli/anyscale/cloud_resource.py#L378
                "FromPort": -1,
                "ToPort": -1,
                "UserIdGroupPairs": [{"GroupId": security_group_id}],
            },
        ],
    )
    return security_group_id


def _create_bucket(
    bucket_name: str = "mock_bucket", cors_rules: Any = None, region: str = "us-east-1"
) -> Any:
    s3 = boto3.client("s3")
    kwargs = {}
    if region != "us-east-1":
        kwargs["CreateBucketConfiguration"] = {"LocationConstraint": region}

    s3.create_bucket(Bucket=bucket_name, **kwargs)  # type: ignore
    if cors_rules:
        s3.put_bucket_cors(
            Bucket=bucket_name, CORSConfiguration={"CORSRules": cors_rules}
        )
    return boto3.resource("s3").Bucket(bucket_name)


def _create_efs(
    region: str,
    subnet_ids: List[str],
    security_group_ids: List[str],
    mount_target_ips: Any = None,
    backup: bool = False,
):
    efs = boto3.client("efs", region_name=region)
    create_resp = efs.create_file_system(CreationToken="mock_efs", Backup=backup)
    efs_id = create_resp["FileSystemId"]

    for subnet_id in subnet_ids:
        if mount_target_ips:
            efs.create_mount_target(
                FileSystemId=efs_id,
                SubnetId=subnet_id,
                SecurityGroups=security_group_ids,
                IpAddress=mount_target_ips[subnet_id],
            )
        else:
            efs.create_mount_target(
                FileSystemId=efs_id,
                SubnetId=subnet_id,
                SecurityGroups=security_group_ids,
            )
    return efs_id


def _create_cloudformation_stack(region: str) -> str:
    cloudformation = boto3.client("cloudformation", region_name=region)
    cfn = cloudformation.create_stack(
        StackName="mock_cloudformation_stack",
        TemplateBody=json.dumps(
            {
                "Resources": {
                    "mock_resource": {
                        "Type": "AWS::EC2::Instance",
                        "Properties": {
                            "InstanceType": "t2.micro",
                            "ImageId": "ami-0b69ea66ff7391e80",
                        },
                    }
                }
            }
        ),
    )

    return cfn["StackId"]


@pytest.mark.parametrize(
    ("vpc_exists", "vpc_cidr_block", "expected_result", "expected_log_message"),
    [
        pytest.param(False, "0.0.0.0/0", False, r"Could not find"),
        # Happy sizes
        pytest.param(True, "0.0.0.0/16", True, r"verification succeeded."),
        pytest.param(True, "0.0.0.0/20", True, r"verification succeeded."),
        # Warn sizes
        pytest.param(
            True, "0.0.3.0/21", True, r"but this VPC only supports up to \d+ addresses",
        ),
        pytest.param(
            True, "0.0.4.0/24", True, r"but this VPC only supports up to \d+ addresses",
        ),
        # Error sizes
        pytest.param(
            True,
            "0.0.4.0/25",
            False,
            r"Please reach out to support if this is an issue!",
        ),
    ],
)
@mock_ec2
def test_verify_aws_vpc(
    capsys,
    vpc_exists: bool,
    vpc_cidr_block: str,
    expected_result: bool,
    expected_log_message: str,
):
    cloud_resource_mock = generate_cloud_resource_mock_aws()
    region = "us-west-2"
    if vpc_exists:
        vpc_id = _create_vpc(cidr_block=vpc_cidr_block, region=region)
        cloud_resource_mock.aws_vpc_id = vpc_id

    result = verify_aws_vpc(
        cloud_resource=cloud_resource_mock,
        boto3_session=boto3.Session(region_name=region),
        logger=CloudSetupLogger(),
    )
    assert result == expected_result

    stdout, stderr = capsys.readouterr()
    sys.stdout.write(stdout)
    sys.stderr.write(stderr)

    if expected_log_message:
        assert re.search(expected_log_message, stderr)


@pytest.mark.parametrize(
    ("is_private_network"), [False, True],
)
@mock_ec2
def test_verify_aws_subnets(capsys, is_private_network: bool):
    cloud_resource_mock = generate_cloud_resource_mock_aws()
    region = "us-west-2"
    vpc_id = _create_vpc(region=region)
    subnet_ids = _create_subnets(
        vpc_id=vpc_id, region=region, assign_public_ip=not is_private_network
    )

    cloud_resource_mock.aws_vpc_id = vpc_id
    cloud_resource_mock.aws_subnet_ids_with_availability_zones = subnet_ids

    result = verify_aws_subnets(
        cloud_resource=cloud_resource_mock,
        region=region,
        is_private_network=is_private_network,
        logger=CloudSetupLogger(),
    )

    _, stderr = capsys.readouterr()

    assert result
    assert not re.search(
        r"does not have the 'Auto-assign Public IP' option enabled", stderr
    )
    assert re.search(r"verification succeeded.", stderr)


@mock_ec2
def test_verify_aws_subnets_warn_about_not_assigning_public_ip_in_public_subnets(
    capsys,
):
    cloud_resource_mock = generate_cloud_resource_mock_aws()
    region = "us-west-2"
    vpc_id = _create_vpc(region=region)
    subnet_ids = _create_subnets(vpc_id=vpc_id, region=region, assign_public_ip=False)

    cloud_resource_mock.aws_vpc_id = vpc_id
    cloud_resource_mock.aws_subnet_ids_with_availability_zones = subnet_ids

    result = verify_aws_subnets(
        cloud_resource=cloud_resource_mock,
        region=region,
        is_private_network=False,
        logger=CloudSetupLogger(),
    )

    _, stderr = capsys.readouterr()

    assert result
    assert re.search(
        r"does not have the 'Auto-assign Public IP' option enabled", stderr
    )
    assert re.search(r"verification succeeded.", stderr)


@mock_ec2
def test_verify_aws_subnets_vpc_subnet_mismatch(capsys):
    cloud_resource_mock = generate_cloud_resource_mock_aws()

    region = "us-west-2"
    vpc_id = _create_vpc(region=region)
    subnet_ids = _create_subnets(vpc_id=vpc_id, region=region, num_subnets=2)

    irrevalent_vpc_id = _create_vpc(region=region)

    cloud_resource_mock.aws_vpc_id = irrevalent_vpc_id
    cloud_resource_mock.aws_subnet_ids_with_availability_zones = subnet_ids

    result = verify_aws_subnets(
        cloud_resource=cloud_resource_mock,
        region=region,
        is_private_network=False,
        logger=CloudSetupLogger(),
    )

    _, stderr = capsys.readouterr()

    assert not result
    assert re.search(r"is not in a vpc of this cloud.", stderr)


@mock_ec2
def test_verify_aws_subnets_insufficient_subnets(capsys):
    cloud_resource_mock = generate_cloud_resource_mock_aws()

    region = "us-west-2"
    vpc_id = _create_vpc(region=region)
    subnet_ids = _create_subnets(vpc_id=vpc_id, region=region, num_subnets=1)

    cloud_resource_mock.aws_vpc_id = vpc_id
    cloud_resource_mock.aws_subnet_ids_with_availability_zones = subnet_ids

    result = verify_aws_subnets(
        cloud_resource=cloud_resource_mock,
        region=region,
        is_private_network=False,
        logger=CloudSetupLogger(),
    )
    assert result is False
    _, stderr = capsys.readouterr()
    assert re.search(r"Need at least 2 subnets for a cloud.", stderr)


@mock_ec2
def test_verify_aws_subnets_insufficient_azs(capsys):
    cloud_resource_mock = generate_cloud_resource_mock_aws()

    region = "us-west-2"
    vpc_id = _create_vpc(region=region)
    subnet_ids = _create_subnets(
        vpc_id=vpc_id, region=region, num_subnets=2, use_single_availability_zone=True
    )

    cloud_resource_mock.aws_vpc_id = vpc_id
    cloud_resource_mock.aws_subnet_ids_with_availability_zones = subnet_ids

    result = verify_aws_subnets(
        cloud_resource=cloud_resource_mock,
        region=region,
        is_private_network=False,
        logger=CloudSetupLogger(),
    )
    assert result is False
    _, stderr = capsys.readouterr()
    assert re.search(r"Subnets should be in at least 2 Availability Zones", stderr)


@mock_ec2
def test_verify_aws_subnets_no_subnets():
    cloud_resource_mock = generate_cloud_resource_mock_aws()

    cloud_resource_mock.aws_subnet_ids_with_availability_zones = None

    result = verify_aws_subnets(
        cloud_resource=cloud_resource_mock,
        region="us-west-2",
        is_private_network=False,
        logger=CloudSetupLogger(),
    )
    assert result is False


@pytest.mark.parametrize(
    ("subnet_cidr", "has_capacity", "expected_log_message"),
    [
        # Happy sizes
        ("10.0.0.0/16", True, None),
        ("0.0.0.0/0", True, None),
        ("0.0.1.0/24", True, None),
        # Warn sizes
        ("0.0.2.0/25", True, r"but this Subnet only supports up to \d+ addresses"),
        ("0.0.3.0/28", True, r"but this Subnet only supports up to \d+ addresses"),
        # Error sizes
        ("0.0.3.0/29", False, r"Please reach out to support if this is an issue!"),
    ],
)
def test_aws_subnet_has_enough_capacity(
    subnet_cidr, has_capacity, expected_log_message, capsys
):
    subnet = Mock()
    subnet.id = "vpc-fake_id"
    subnet.cidr_block = subnet_cidr

    assert (
        AWS_SUBNET_CAPACITY.verify_network_capacity(
            cidr_block_str=subnet_cidr,
            resource_name="Subnet",
            logger=CloudSetupLogger(),
        )
        == has_capacity
    )

    stdout, stderr = capsys.readouterr()
    sys.stdout.write(stdout)
    sys.stderr.write(stderr)

    if expected_log_message:
        assert re.search(expected_log_message, stderr)


@mock_iam
@pytest.mark.parametrize(
    ("iam_roles_mock_fn", "expect_warning"),
    [
        pytest.param(lambda _: None, True, id="No Policies"),
        pytest.param(_attach_iam_roles_to_role, False, id="Separate Policies"),
        pytest.param(_attach_inline_iam_policy_to_role, False, id="Inline Policies"),
    ],
)
def test_verify_aws_iam_roles(
    iam_roles_mock_fn: Callable[[str], None], expect_warning: bool
):
    cloud_resource_mock = generate_cloud_resource_mock_aws()
    cp_role_arn, dp_role_arn = _create_roles("fake_anyscale_aws_account")
    _create_instance_profile_for_role(dp_role_arn, None)
    cloud_resource_mock.aws_iam_role_arns = [cp_role_arn, dp_role_arn]
    iam_roles_mock_fn(cp_role_arn)
    mock_logger = Mock()
    assert verify_aws_iam_roles(
        cloud_resource=cloud_resource_mock,
        boto3_session=boto3.Session(),
        anyscale_aws_account="fake_anyscale_aws_account",
        logger=mock_logger,
        cloud_id="fake_cloud_id",
    )

    if expect_warning:
        assert len(mock_logger.warning.mock_calls) == 1
    else:
        mock_logger.warning.assert_not_called()

    assert len(mock_logger.info.mock_calls) == 2


@mock_iam
@pytest.mark.parametrize("include_service_permissions", [True, False])
def test_verify_aws_iam_service_permissions(include_service_permissions: bool,):
    """Test service permission validation
    """
    cloud_resource_mock = generate_cloud_resource_mock_aws()
    cp_role_arn, dp_role_arn = _create_roles("fake_anyscale_aws_account")
    _create_instance_profile_for_role(dp_role_arn, None)
    cloud_resource_mock.aws_iam_role_arns = [cp_role_arn, dp_role_arn]

    iam = boto3.client("iam")
    role_name = AwsRoleArn.from_string(cp_role_arn).to_role_name()

    iam.put_role_policy(
        RoleName=role_name,
        PolicyName="".join(random.choices(string.ascii_lowercase, k=10)),
        PolicyDocument=json.dumps(ANYSCALE_IAM_PERMISSIONS_EC2_STEADY_STATE),
    )
    if include_service_permissions:
        iam.put_role_policy(
            RoleName=role_name,
            PolicyName="".join(random.choices(string.ascii_lowercase, k=10)),
            PolicyDocument=json.dumps(ANYSCALE_IAM_PERMISSIONS_SERVICE_STEADY_STATE),
        )
    iam.put_role_policy(
        RoleName=role_name,
        PolicyName="".join(random.choices(string.ascii_lowercase, k=10)),
        PolicyDocument=json.dumps(ANYSCALE_IAM_PERMISSIONS_EC2_INITIAL_RUN),
    )

    mock_logger = Mock()
    test_status = verify_aws_iam_roles(
        cloud_resource=cloud_resource_mock,
        boto3_session=boto3.Session(),
        anyscale_aws_account="fake_anyscale_aws_account",
        logger=mock_logger,
        cloud_id="fake_cloud_id",
    )

    assert test_status

    if include_service_permissions:

        mock_logger.warning.assert_not_called()
        assert len(mock_logger.info.mock_calls) == 2

    else:
        assert len(mock_logger.confirm_missing_permission.mock_calls) == 1


@mock_iam
@pytest.mark.parametrize(
    ("cp_role_exist", "dp_role_exist"),
    [pytest.param(True, False), pytest.param(False, True),],
)
def test_verify_aws_iam_roles_not_exist(cp_role_exist: bool, dp_role_exist: bool):
    cloud_resource_mock = generate_cloud_resource_mock_aws()
    cp_role_arn, dp_role_arn = _create_roles("fake_anyscale_aws_account")

    if not cp_role_exist:
        _delete_role(cp_role_arn)

    if not dp_role_exist:
        _delete_role(dp_role_arn)

    assert not verify_aws_iam_roles(
        cloud_resource=cloud_resource_mock,
        boto3_session=boto3.Session(),
        anyscale_aws_account="fake_anyscale_aws_account",
        logger=CloudSetupLogger(),
        cloud_id="fake_cloud_id",
    )


@mock_iam
def test_verify_aws_iam_data_plane_role_trust_policy():
    """Tests that the dataplane role only passes validation if it
    gives trust to AWS service EC2."""
    cloud_resource_mock = generate_cloud_resource_mock_aws()
    cp_role_arn, dp_role_arn = _create_roles("fake_anyscale_aws_account")
    cloud_resource_mock.aws_iam_role_arns = [cp_role_arn, dp_role_arn]
    mock_logger = Mock()
    session = boto3.Session()
    print(
        verify_aws_iam_roles(
            cloud_resource=cloud_resource_mock,
            boto3_session=session,
            anyscale_aws_account="fake_anyscale_aws_account",
            logger=mock_logger,
            cloud_id="fake_cloud_id",
        )
    )

    iam = boto3.client("iam")
    suffix = "".join(random.choices(string.ascii_lowercase, k=10))
    dataplane_role_name = DEFAULT_RAY_IAM_ROLE + suffix
    # create a dataplane role without assume role policy document:
    dp_role_arn = iam.create_role(
        RoleName=dataplane_role_name, AssumeRolePolicyDocument="{}"
    )["Role"]["Arn"]

    iam.attach_role_policy(
        RoleName=dataplane_role_name, PolicyArn=AMAZON_ECR_READONLY_ACCESS_POLICY_ARN
    )
    iam.attach_role_policy(
        RoleName=dataplane_role_name, PolicyArn=AMAZON_S3_FULL_ACCESS_POLICY_ARN
    )
    cloud_resource_mock.aws_iam_role_arns = [cp_role_arn, dp_role_arn]

    # No AssumeRolePolicyDocument for role
    assert not verify_aws_iam_roles(
        cloud_resource=cloud_resource_mock,
        boto3_session=session,
        anyscale_aws_account="fake_anyscale_aws_account",
        logger=mock_logger,
        cloud_id="fake_cloud_id",
    )


@mock_iam
def test_verify_aws_iam_roles_instance_profiles():
    """Tests that the dataplane role only passes validation if it has
    an InstanceProfile with the same name as the Role."""
    cloud_resource_mock = generate_cloud_resource_mock_aws()
    cp_role_arn, dp_role_arn = _create_roles("fake_anyscale_aws_account")
    cloud_resource_mock.aws_iam_role_arns = [cp_role_arn, dp_role_arn]
    mock_logger = Mock()

    # No instance profile for role
    assert not verify_aws_iam_roles(
        cloud_resource=cloud_resource_mock,
        boto3_session=boto3.Session(),
        anyscale_aws_account="fake_anyscale_aws_account",
        logger=mock_logger,
        cloud_id="fake_cloud_id",
    )

    # Create an instance profile for the role, but with the wrong name
    _create_instance_profile_for_role(dp_role_arn, "wrong_name")
    assert not verify_aws_iam_roles(
        cloud_resource=cloud_resource_mock,
        boto3_session=boto3.Session(),
        anyscale_aws_account="fake_anyscale_aws_account",
        logger=mock_logger,
        cloud_id="fake_cloud_id",
    )

    # Create a correct insance profile with the role!
    _create_instance_profile_for_role(dp_role_arn, None)
    assert verify_aws_iam_roles(
        cloud_resource=cloud_resource_mock,
        boto3_session=boto3.Session(),
        anyscale_aws_account="fake_anyscale_aws_account",
        logger=mock_logger,
        cloud_id="fake_cloud_id",
    )


@mock_iam
def test_verify_aws_iam_roles_different_account(capsys):
    """Tests that the IAM roles must be in the same account."""
    cloud_resource_mock = generate_cloud_resource_mock_aws()
    cp_role_arn, _ = _create_roles("fake_anyscale_aws_account")
    with patch.dict(os.environ, {"MOTO_ACCOUNT_ID": "112233445566"}):
        _, dp_role_arn = _create_roles("fake_anyscale_aws_account")
    cloud_resource_mock.aws_iam_role_arns = [cp_role_arn, dp_role_arn]

    assert not verify_aws_iam_roles(
        cloud_resource=cloud_resource_mock,
        boto3_session=boto3.Session(),
        anyscale_aws_account="fake_anyscale_aws_account",
        logger=CloudSetupLogger(),
        cloud_id="fake_cloud_id",
    )

    _, stderr = capsys.readouterr()
    assert "All IAM roles must be in the same AWS account:" in stderr


@pytest.mark.parametrize(
    ("sg_list", "expected_result", "expected_log_message"),
    [
        pytest.param([], False, r"Missing security group IDs", id="SG-Does-Not-Exist"),
        pytest.param([False], False, r"Could not find", id="SG-Exists-Not-Valid"),
        pytest.param([True], True, None, id="SG-Exists-And-Valid"),
        pytest.param(
            [True, False], False, r"Could not find", id="2SG-Exists-Not-Valid"
        ),
        pytest.param([True, True], True, None, id="2SG-Exists-And-Valid"),
    ],
)
@mock_ec2
def test_verify_aws_security_groups_exist_and_valid(
    capsys, sg_list: List[bool], expected_result: bool, expected_log_message: str,
):
    cloud_resource_mock = generate_cloud_resource_mock_aws()

    region = "us-west-2"
    cloud_resource_mock.aws_security_groups = []
    for sg_valid in sg_list:
        if sg_valid:  # True means valid sg, False throws not found error
            sg_id = _create_security_group(region)
        else:
            sg_id = "Invalid_SG"
        cloud_resource_mock.aws_security_groups.append(sg_id)

    result = verify_aws_security_groups(
        cloud_resource=cloud_resource_mock,
        boto3_session=boto3.Session(region_name=region),
        logger=CloudSetupLogger(),
    )

    stdout, stderr = capsys.readouterr()
    sys.stdout.write(stdout)
    sys.stderr.write(stderr)

    assert result == expected_result
    if expected_log_message:
        assert re.search(expected_log_message, stderr)


@pytest.mark.parametrize(
    ("not_enough_permission", "excessive_permission", "expected_log_message"),
    [
        pytest.param(False, False, None, id="Expected_Permission"),
        pytest.param(
            True,
            False,
            r"not contain inbound permission for ports",
            id="Not_Enough_Permission",
        ),
        pytest.param(
            False, True, r"allows access to more than", id="Excessive_Permission"
        ),
    ],
)
@mock_ec2
def test_verify_aws_security_groups_inbound_permissions(
    capsys,
    not_enough_permission: bool,
    excessive_permission: bool,
    expected_log_message: str,
):
    cloud_resource_mock = generate_cloud_resource_mock_aws()

    region = "us-west-2"

    if not_enough_permission:
        sg_id = _create_security_group(region)
    elif excessive_permission:
        sg_id = _create_security_group_inbound_permission_excessive_permission(region)
    else:  # expected permission
        sg_id = _create_security_group_inbound_permission_expected(region)

    cloud_resource_mock.aws_security_groups = [sg_id]

    result = verify_aws_security_groups(
        cloud_resource=cloud_resource_mock,
        boto3_session=boto3.Session(region_name=region),
        logger=CloudSetupLogger(),
    )
    stdout, stderr = capsys.readouterr()
    sys.stdout.write(stdout)
    sys.stderr.write(stderr)

    assert result
    if expected_log_message:
        assert re.search(expected_log_message, stderr)


@pytest.mark.parametrize(
    ("security_groups_permissions", "expected_result", "expected_log_message"),
    [
        pytest.param(
            [["https"], ["ssh"], ["self"]], True, None, id="Expected_Permission"
        ),
        pytest.param(
            [["https", "ssh", "self"], ["ssh", "self"], ["https"]],
            True,
            None,
            id="Expected_Permission_With_Duplicates",
        ),
        pytest.param(
            [["https", "self"], ["self"]],
            True,
            r"not contain inbound permission for ports",
            id="Missing_Permission",
        ),
        pytest.param(
            [["https", "ssh"], ["https"]],
            False,
            r"not contain inbound permission for all ports for traffic from the same",
            id="Missing_Self_Permission",
        ),
        pytest.param(
            [["https", "excess"], ["ssh"], ["self"]],
            True,
            r"allows access to more than",
            id="Excessive_Permission",
        ),
    ],
)
@mock_ec2
def test_verify_aws_security_groups_multiple_security_groups(
    capsys,
    security_groups_permissions: List[List[str]],
    expected_result,
    expected_log_message: str,
):
    """Tests permission check is correct for multiple security groups"""
    cloud_resource_mock = generate_cloud_resource_mock_aws()

    region = "us-west-2"
    sg_ids = []

    for security_group_permission in security_groups_permissions:
        sg_id = _create_security_group_given_permissions(
            security_group_permission, region
        )
        sg_ids.append(sg_id)

    cloud_resource_mock.aws_security_groups = sg_ids

    result = verify_aws_security_groups(
        cloud_resource=cloud_resource_mock,
        boto3_session=boto3.Session(region_name=region),
        logger=CloudSetupLogger(),
    )
    stdout, stderr = capsys.readouterr()
    sys.stdout.write(stdout)
    sys.stderr.write(stderr)

    assert result == expected_result
    if expected_log_message:
        assert re.search(expected_log_message, stderr)


@pytest.mark.parametrize(
    ("s3_exists", "expected_result"),
    [
        pytest.param(False, False, id="S3-Does-Not-Exist"),
        pytest.param(True, True, id="S3-Exists"),
    ],
)
@pytest.mark.parametrize(
    ("cors_rules", "expected_warning_call_count"),
    [
        pytest.param(
            [
                {
                    "AllowedHeaders": ["*"],
                    "AllowedMethods": ["GET"],
                    "AllowedOrigins": [ANYSCALE_HOST],
                    "ExposeHeaders": [],
                }
            ],
            0,
            id="CorrectCORS",
        ),
        pytest.param(
            [
                {
                    "AllowedHeaders": ["*"],
                    "AllowedMethods": ["GET"],
                    "AllowedOrigins": ["openai.com"],
                    "ExposeHeaders": [],
                }
            ],
            1,
            id="ImproperCORS",
        ),
        pytest.param([], 2, id="NoCORS"),
    ],
)
@mock_s3
@mock_iam
def test_verify_aws_s3(
    s3_exists: bool,
    expected_result: bool,
    cors_rules: List[Dict[str, List[str]]],
    expected_warning_call_count: int,
):
    cloud_resource_mock = generate_cloud_resource_mock_aws()

    region = "us-west-2"
    if s3_exists:
        bucket = _create_bucket(cors_rules=cors_rules, region=region)
        cloud_resource_mock.aws_s3_id = bucket.name

    cp_role_arn, dp_role_arn = _create_roles("fake_anyscale_aws_account")

    cloud_resource_mock.aws_iam_role_arns = [cp_role_arn, dp_role_arn]

    mock_logger = Mock()

    with patch.multiple(
        "anyscale.cloud_resource", verify_s3_access=Mock(return_value=True),
    ):
        result = verify_aws_s3(
            cloud_resource=cloud_resource_mock,
            boto3_session=boto3.Session(region_name=region),
            region=region,
            logger=mock_logger,
        )

    if s3_exists:
        assert mock_logger.warning.call_count == expected_warning_call_count
    else:
        assert mock_logger.error.call_count == 1
        assert mock_logger.warning.call_count == 0

    assert result == expected_result


@mock_s3
@mock_iam
def test_verify_aws_s3_wrong_region(capsys):
    cloud_resource_mock = generate_cloud_resource_mock_aws()

    region = "us-west-2"
    bucket = _create_bucket(region="us-east-1")
    cloud_resource_mock.aws_s3_id = bucket.name

    cp_role_arn, dp_role_arn = _create_roles("fake_anyscale_aws_account")
    cloud_resource_mock.aws_iam_role_arns = [cp_role_arn, dp_role_arn]

    with patch.multiple(
        "anyscale.cloud_resource", verify_s3_access=Mock(return_value=True),
    ):
        verify_aws_s3(
            cloud_resource=cloud_resource_mock,
            boto3_session=boto3.Session(region_name=region),
            region=region,
            logger=CloudSetupLogger(),
        )

    _, stderr = capsys.readouterr()

    assert re.search("S3 bucket .* is in region us-east-1, .* us-west-2", stderr)


@pytest.mark.parametrize(
    (
        "efs_exists",
        "num_subnets_to_use",
        "use_security_group",
        "expected_result",
        "warning",
        "enable_backup",
    ),
    [
        pytest.param(False, 2, True, False, False, True),
        pytest.param(True, 0, False, False, False, True),
        pytest.param(True, 1, True, True, True, True),
        pytest.param(True, 1, False, True, True, True),
        pytest.param(True, 2, True, True, False, True),
        pytest.param(True, 2, False, True, True, True),
        pytest.param(True, 2, False, True, True, False),
    ],
)
@mock_efs
@mock_ec2
def test_verify_aws_efs(
    capsys,
    efs_exists: bool,
    num_subnets_to_use: int,
    use_security_group: bool,
    expected_result: bool,
    warning: bool,
    enable_backup: bool,
):
    cloud_resource_mock = generate_cloud_resource_mock_aws()

    region = "us-west-2"
    if efs_exists:
        vpc_id = _create_vpc(region=region)
        aws_subnet_ids = _create_subnets(vpc_id, region, num_subnets=2)
        assert len(aws_subnet_ids) == 2
        security_group_id = _create_security_group(region)
        subnet_ids_in_efs = [
            subnet.subnet_id for subnet in aws_subnet_ids[:num_subnets_to_use]
        ]
        security_groups_in_efs = [security_group_id] if use_security_group else []
        efs_id = _create_efs(
            region, subnet_ids_in_efs, security_groups_in_efs, backup=enable_backup
        )
        cloud_resource_mock.aws_subnet_ids_with_availability_zones = aws_subnet_ids
        cloud_resource_mock.aws_security_groups = [security_group_id]
        cloud_resource_mock.aws_efs_id = efs_id

    result = verify_aws_efs(
        cloud_resource=cloud_resource_mock,
        boto3_session=boto3.Session(region_name=region),
        logger=CloudSetupLogger(),
    )
    _, stderr = capsys.readouterr()
    if warning:
        assert re.search(r"does not contain a mount target with the subnet", stderr)
    if enable_backup:
        assert not re.search(r"backup policy is not enabled", stderr)
        assert not re.search(r"backup policy not found", stderr)
    else:
        assert re.search(r"backup policy is not enabled", stderr) or re.search(
            r"backup policy not found", stderr
        )
    assert result == expected_result


@pytest.mark.parametrize(
    ("valid_ip", "expected_result", "expected_log_message",),
    [
        pytest.param(
            False, False, "Mount target registered with the cloud no longer exists."
        ),
        pytest.param(True, True, None),
    ],
)
@mock_efs
@mock_ec2
def test_verify_aws_efs_valid_ip(
    capsys, valid_ip: bool, expected_result: bool, expected_log_message: str,
):
    cloud_resource_mock = generate_cloud_resource_mock_aws()

    region = "us-west-2"

    vpc_id = _create_vpc(region=region)
    subnet_ids_with_az = _create_subnets(vpc_id, region, num_subnets=2)
    subnet_ids = [
        subnet_id_with_az.subnet_id for subnet_id_with_az in subnet_ids_with_az
    ]
    security_groups = [_create_security_group(region)]

    mount_target_ips = {
        subnet_id_with_az.subnet_id: f"10.0.{i}.123"
        for i, subnet_id_with_az in enumerate(subnet_ids_with_az)
    }

    efs_id = _create_efs(
        region, subnet_ids, security_groups, mount_target_ips, backup=True
    )
    cloud_resource_mock.aws_subnet_ids = subnet_ids
    cloud_resource_mock.aws_security_groups = security_groups
    cloud_resource_mock.aws_efs_id = efs_id

    if valid_ip:
        cloud_resource_mock.aws_efs_mount_target_ip = mount_target_ips[subnet_ids[0]]
    else:
        cloud_resource_mock.aws_efs_mount_target_ip = "1.2.3.4"

    result = verify_aws_efs(
        cloud_resource=cloud_resource_mock,
        boto3_session=boto3.Session(region_name=region),
        logger=CloudSetupLogger(),
    )

    stdout, stderr = capsys.readouterr()
    sys.stdout.write(stdout)
    sys.stderr.write(stderr)

    assert result == expected_result
    if expected_log_message:
        assert re.search(expected_log_message, stderr)


@pytest.mark.parametrize(
    ("cloudformation_stack_exists", "expected_result"),
    [pytest.param(False, False), pytest.param(True, True)],
)
@mock_cloudformation
def test_verify_aws_cloudformation_stack(
    cloudformation_stack_exists: bool, expected_result: bool
):
    cloud_resource_mock = generate_cloud_resource_mock_aws()

    region = "us-west-2"
    if cloudformation_stack_exists:
        cloud_resource_mock.aws_cloudformation_stack_id = _create_cloudformation_stack(
            region
        )

    result = verify_aws_cloudformation_stack(
        cloud_resource=cloud_resource_mock,
        boto3_session=boto3.Session(region_name=region),
        logger=CloudSetupLogger(),
    )
    assert result == expected_result


@pytest.mark.parametrize(
    (
        "enable_head_node_fault_tolerance",
        "redis_cluster_exist",
        "matching_security_group",
        "matching_vpc",
        "matching_subnets",
        "has_lru_set",
        "should_raise_exception",
        "expected_result",
        "expected_log_message",
    ),
    [
        pytest.param(
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            "Missing MemoryDB cluster id",
        ),
        pytest.param(
            True,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            "Could not find MemoryDB cluster",
        ),
        pytest.param(
            True,
            True,
            False,
            False,
            False,
            False,
            False,
            False,
            "that is not the same as the cloud's security group",
        ),
        pytest.param(
            True,
            True,
            True,
            False,
            False,
            False,
            False,
            False,
            "is not in the same VPC as the cloud",
        ),
        pytest.param(
            True,
            True,
            True,
            True,
            False,
            False,
            False,
            False,
            "that is not one of the subnets in the cloud",
        ),
        pytest.param(
            True,
            True,
            True,
            True,
            True,
            False,
            False,
            False,
            "should have parameter group with maxmemory-policy set to allkeys-lru",
        ),
        pytest.param(
            True,
            True,
            True,
            True,
            True,
            True,
            True,
            False,
            "Failed to verify MemoryDB cluster",
        ),
        pytest.param(True, True, True, True, True, True, False, True, None),
    ],
)
def test_verify_aws_memorydb_cluster(  # noqa: PLR0913
    capsys,
    enable_head_node_fault_tolerance,
    redis_cluster_exist,
    matching_security_group,
    matching_vpc,
    matching_subnets,
    has_lru_set,
    should_raise_exception,
    expected_result,
    expected_log_message,
):
    cloud_resource_mock = generate_cloud_resource_mock_aws(
        enable_head_node_fault_tolerance=enable_head_node_fault_tolerance
    )

    if not redis_cluster_exist:
        mock_cluster_resp = {"Clusters": []}
    else:
        mock_cluster_resp = {
            "Clusters": [
                {
                    "SecurityGroups": [
                        {
                            "SecurityGroupId": cloud_resource_mock.aws_security_groups[
                                0
                            ]
                            if matching_security_group
                            else "not-matching-security-group"
                        }
                    ],
                    "SubnetGroupName": "subnet-group-1234567890",
                    "ParameterGroupName": "parameter-group-1234567890",
                },
            ]
        }

    mock_subnet_group_resp = {
        "SubnetGroups": [
            {
                "SubnetGroupName": "subnet-group-1234567890",
                "VpcId": cloud_resource_mock.aws_vpc_id
                if matching_vpc
                else "not-matching-vpc",
            }
        ]
    }

    if matching_subnets:
        mock_subnet_group_resp["SubnetGroups"][0]["Subnets"] = [
            {
                "Identifier": cloud_resource_mock.aws_subnet_ids_with_availability_zones[
                    0
                ].subnet_id,
            },
            {
                "Identifier": cloud_resource_mock.aws_subnet_ids_with_availability_zones[
                    1
                ].subnet_id,
            },
        ]
    else:
        mock_subnet_group_resp["SubnetGroups"][0]["Subnets"] = [
            {"Identifier": "not-matching-subnet-1",},
            {"Identifier": "not-matching-subnet-2",},
        ]

    if has_lru_set:
        mock_parameter_group_resp = {
            "Parameters": [{"Name": "maxmemory-policy", "Value": "allkeys-lru",}]
        }
    else:
        mock_parameter_group_resp = {
            "Parameters": [{"Name": "maxmemory-policy", "Value": "not-allkeys-lru",}]
        }

    mock_memorydb_client = Mock(
        describe_clusters=Mock(return_value=mock_cluster_resp),
        describe_subnet_groups=Mock(return_value=mock_subnet_group_resp),
        describe_parameters=Mock(return_value=mock_parameter_group_resp),
    )
    if should_raise_exception:
        mock_memorydb_client.describe_clusters.side_effect = ClientError(
            {
                "Error": {
                    "Code": "AccessDeniedException",
                    "Message": "Permission denied",
                }
            },
            "describe_clusters",
        )

    mock_boto3 = Mock(client=Mock(return_value=mock_memorydb_client))

    if should_raise_exception:
        with pytest.raises(ClickException, match=expected_log_message):
            verify_aws_memorydb_cluster(
                cloud_resource=cloud_resource_mock,
                boto3_session=mock_boto3,
                logger=CloudSetupLogger(),
                strict=True,
            )
        return

    result = verify_aws_memorydb_cluster(
        cloud_resource=cloud_resource_mock,
        boto3_session=mock_boto3,
        logger=CloudSetupLogger(),
        strict=True,
    )

    stdout, stderr = capsys.readouterr()
    sys.stdout.write(stdout)
    sys.stderr.write(stderr)

    assert result == expected_result
    if expected_log_message:
        assert re.search(expected_log_message, stderr)



from typing import Any, Dict


# Used for data-gplane role.
AMAZON_S3_FULL_ACCESS_POLICY_NAME = "AmazonS3FullAccess"
AMAZON_S3_FULL_ACCESS_POLICY_ARN = (
    f"arn:aws:iam::aws:policy/{AMAZON_S3_FULL_ACCESS_POLICY_NAME}"
)

AMAZON_ECR_READONLY_ACCESS_POLICY_NAME = "AmazonEC2ContainerRegistryReadOnly"
AMAZON_ECR_READONLY_ACCESS_POLICY_ARN = (
    f"arn:aws:iam::aws:policy/{AMAZON_ECR_READONLY_ACCESS_POLICY_NAME}"
)
DEFAULT_RAY_IAM_ASSUME_ROLE_POLICY = {
    "Version": "2012-10-17",
    "Statement": {
        "Effect": "Allow",
        "Principal": {"Service": ["ec2.amazonaws.com"]},
        "Action": "sts:AssumeRole",
    },
}

# Used for control-plane role.
ANYSCALE_IAM_POLICY_NAME_STEADY_STATE = "Anyscale_IAM_Policy_Steady_State"

# Refer to https://docs.anyscale.com/cloud-deployment/aws/manage-clouds#appendix-detailed-resource-requirements
ANYSCALE_IAM_PERMISSIONS_EC2_STEADY_STATE: Dict[str, Any] = {
    "Version": "2012-10-17",
    "Statement": [
        {
            "Sid": "IAM",
            "Effect": "Allow",
            "Action": ["iam:PassRole", "iam:GetInstanceProfile"],
            "Resource": "*",
        },
        {
            "Sid": "RetrieveGenericAWSResources",
            "Effect": "Allow",
            "Action": [
                # Populates metadata about what is available
                # in the account.
                "ec2:DescribeAvailabilityZones",
                "ec2:DescribeInstanceTypes",
                "ec2:DescribeRegions",
                "ec2:DescribeAccountAttributes",
            ],
            "Resource": "*",
        },
        {
            "Sid": "DescribeRunningResources",
            "Effect": "Allow",
            "Action": [
                # Determines cluster/configuration status.
                "ec2:DescribeInstances",
                "ec2:DescribeSubnets",
                "ec2:DescribeRouteTables",
                "ec2:DescribeSecurityGroups",
            ],
            "Resource": "*",
        },
        {
            "Sid": "InstanceManagementCore",
            "Effect": "Allow",
            "Action": [
                # Minimal Permissions to Run Instances on Anyscale.
                "ec2:RunInstances",
                "ec2:StartInstances",
                "ec2:StopInstances",
                "ec2:TerminateInstances",
            ],
            "Resource": "*",
        },
        {
            "Sid": "InstanceTagMangement",
            "Effect": "Allow",
            "Action": ["ec2:CreateTags", "ec2:DeleteTags"],
            "Resource": "*",
        },
        {
            "Sid": "InstanceManagementSpot",
            "Effect": "Allow",
            "Action": [
                # Extended Permissions to Run Instances on Anyscale.
                "ec2:CancelSpotInstanceRequests",
                "ec2:ModifyImageAttribute",
                "ec2:ModifyInstanceAttribute",
                "ec2:RequestSpotInstances",
            ],
            "Resource": "*",
        },
        {
            "Sid": "ResourceManagementExtended",
            "Effect": "Allow",
            "Action": [
                # Volume management
                "ec2:AttachVolume",
                "ec2:CreateVolume",
                "ec2:DescribeVolumes",
                # IAMInstanceProfiles
                "ec2:AssociateIamInstanceProfile",
                "ec2:DisassociateIamInstanceProfile",
                "ec2:ReplaceIamInstanceProfileAssociation",
                # Placement groups
                "ec2:CreatePlacementGroup",
                # Address Management
                "ec2:AllocateAddress",
                "ec2:ReleaseAddress",
                # Additional DescribeResources
                "ec2:DescribeIamInstanceProfileAssociations",
                "ec2:DescribeInstanceStatus",
                "ec2:DescribePlacementGroups",
                "ec2:DescribePrefixLists",
                "ec2:DescribeReservedInstancesOfferings",
                "ec2:DescribeSpotInstanceRequests",
                "ec2:DescribeSpotPriceHistory",
            ],
            "Resource": "*",
        },
        {
            "Action": ["elasticfilesystem:DescribeMountTargets"],
            "Effect": "Allow",
            "Resource": "*",
            "Sid": "EFSManagement",
        },
        {
            "Sid": "CreateSpotRole",
            "Effect": "Allow",
            "Action": ["iam:CreateServiceLinkedRole"],
            "Resource": "arn:aws:iam::*:role/aws-service-role/spot.amazonaws.com/AWSServiceRoleForEC2Spot",
            "Condition": {"StringLike": {"iam:AWSServiceName": "spot.amazonaws.com"}},
        },
    ],
}


def get_anyscale_iam_permissions_ec2_restricted(cloud_id: str) -> Dict[str, Any]:
    # Refer to https://docs.anyscale.com/cloud-deployment/aws/manage-clouds#appendix-detailed-resource-requirements
    return {
        "Version": ANYSCALE_IAM_PERMISSIONS_EC2_STEADY_STATE["Version"],
        "Statement": [
            statement
            for statement in ANYSCALE_IAM_PERMISSIONS_EC2_STEADY_STATE["Statement"]
            if statement["Sid"] != "InstanceManagementCore"
        ]
        + [
            {
                "Sid": "DenyTaggingOnOtherInstances",
                "Effect": "Deny",
                "Action": ["ec2:DeleteTags", "ec2:CreateTags"],
                "Resource": "arn:aws:ec2:*:*:instance/*",
                "Condition": {
                    "StringNotEquals": {
                        "aws:ResourceTag/anyscale-cloud-id": f"{cloud_id}",
                        "ec2:CreateAction": ["RunInstances", "StartInstances"],
                    }
                },
            },
            {
                "Sid": "RestrictedEc2Termination",
                "Effect": "Allow",
                "Action": ["ec2:TerminateInstances", "ec2:StopInstances"],
                "Resource": "*",
                "Condition": {
                    "StringEquals": {"aws:ResourceTag/anyscale-cloud-id": f"{cloud_id}"}
                },
            },
            {
                "Sid": "RestrictedInstanceStart",
                "Effect": "Allow",
                "Action": ["ec2:StartInstances", "ec2:RunInstances"],
                "Resource": "*",
                "Condition": {
                    "StringEquals": {"aws:RequestTag/anyscale-cloud-id": f"{cloud_id}"},
                    "ForAnyValue:StringEquals": {"aws:TagKeys": ["anyscale-cloud-id"]},
                },
            },
            {
                "Sid": "AllowRunInstancesForUntaggedResources",
                "Effect": "Allow",
                "Action": "ec2:RunInstances",
                "Resource": [
                    "arn:aws:ec2:*::image/*",
                    "arn:aws:ec2:*::snapshot/*",
                    "arn:aws:ec2:*:*:subnet/*",
                    "arn:aws:ec2:*:*:network-interface/*",
                    "arn:aws:ec2:*:*:security-group/*",
                    "arn:aws:ec2:*:*:key-pair/*",
                    "arn:aws:ec2:*:*:volume/*",
                ],
            },
        ],
    }


ANYSCALE_IAM_POLICY_NAME_SERVICE_STEADY_STATE = (
    "Anyscale_IAM_Policy_Service_Steady_State"
)
ANYSCALE_IAM_PERMISSIONS_SERVICE_STEADY_STATE = {
    "Version": "2012-10-17",
    "Statement": [
        {
            "Sid": "CFN",
            "Effect": "Allow",
            "Action": [
                # CRUD for stack resources
                "cloudformation:TagResource",
                "cloudformation:UntagResource",
                "cloudformation:CreateStack",
                "cloudformation:UpdateStack",
                "cloudformation:DeleteStack",
                "cloudformation:DescribeStackEvents",
                "cloudformation:DescribeStackResources",
                "cloudformation:DescribeStacks",
            ],
            "Resource": "*",
        },
        {
            "Sid": "EC2Describe",
            "Action": ["ec2:DescribeVpcs", "ec2:DescribeInternetGateways"],
            "Effect": "Allow",
            "Resource": "*",
        },
        {
            "Sid": "ELBDescribe",
            "Effect": "Allow",
            "Action": [
                # Describe ELB state information
                "elasticloadbalancing:DescribeListeners",
                "elasticloadbalancing:DescribeLoadBalancers",
                "elasticloadbalancing:DescribeLoadBalancerAttributes",
                "elasticloadbalancing:DescribeRules",
                "elasticloadbalancing:DescribeTargetGroups",
                "elasticloadbalancing:DescribeTargetGroupAttributes",
                "elasticloadbalancing:DescribeTargetHealth",
                "elasticloadbalancing:DescribeListenerCertificates",
            ],
            "Resource": "*",
        },
        {
            "Sid": "ELBCerts",
            "Effect": "Allow",
            "Action": [
                # Add/Remove ELB listener certs
                "elasticloadbalancing:AddListenerCertificates",
                "elasticloadbalancing:RemoveListenerCertificates",
            ],
            "Resource": "*",
        },
        {
            "Sid": "ACMList",
            "Effect": "Allow",
            "Action": [
                # List ELB certs
                "acm:ListCertificates"
            ],
            "Resource": "*",
        },
        {
            "Sid": "ACM",
            "Effect": "Allow",
            "Action": [
                # Manage ACM
                "acm:DeleteCertificate",
                "acm:RenewCertificate",
                "acm:RequestCertificate",
                "acm:AddTagsToCertificate",
                "acm:DescribeCertificate",
                "acm:GetCertificate",
                "acm:ListTagsForCertificate",
            ],
            "Resource": "*",
        },
        {
            "Sid": "ELBWrite",
            "Effect": "Allow",
            "Action": [
                # Modify ELB
                "elasticloadbalancing:AddTags",
                "elasticloadbalancing:RemoveTags",
                "elasticloadbalancing:CreateRule",
                "elasticloadbalancing:ModifyRule",
                "elasticloadbalancing:DeleteRule",
                "elasticloadbalancing:SetRulePriorities",
                "elasticloadbalancing:CreateListener",
                "elasticloadbalancing:ModifyListener",
                "elasticloadbalancing:DeleteListener",
                "elasticloadbalancing:CreateLoadBalancer",
                "elasticloadbalancing:DeleteLoadBalancer",
                "elasticloadbalancing:ModifyLoadBalancerAttributes",
                "elasticloadbalancing:CreateTargetGroup",
                "elasticloadbalancing:ModifyTargetGroup",
                "elasticloadbalancing:DeleteTargetGroup",
                "elasticloadbalancing:ModifyTargetGroupAttributes",
                "elasticloadbalancing:RegisterTargets",
                "elasticloadbalancing:DeregisterTargets",
                "elasticloadbalancing:SetIpAddressType",
                "elasticloadbalancing:SetSecurityGroups",
                "elasticloadbalancing:SetSubnets",
            ],
            "Resource": "*",
            "Condition": {
                "StringEquals": {"aws:CalledViaFirst": "cloudformation.amazonaws.com"}
            },
        },
        {
            "Sid": "CreateELBServiceLinkedRole",
            "Effect": "Allow",
            "Action": "iam:CreateServiceLinkedRole",
            "Resource": "*",
            "Condition": {
                "StringLike": {
                    "iam:AWSServiceName": "elasticloadbalancing.amazonaws.com"
                }
            },
        },
    ],
}

ANYSCALE_IAM_POLICY_NAME_INITIAL_RUN = "Anyscale_IAM_Policy_Initial_Setup"
ANYSCALE_IAM_PERMISSIONS_EC2_INITIAL_RUN = {
    "Version": "2012-10-17",
    "Statement": [
        {
            "Sid": "SetupEC2",
            "Effect": "Allow",
            "Action": [
                # Anyscale runs this on the first time a cloud is configured.
                "ec2:CreateSecurityGroup",
                "ec2:AuthorizeSecurityGroupIngress",
                "ec2:AuthorizeSecurityGroupEgress",
                # For configuring VPCs
                "ec2:DescribeVpcs",
                "ec2:CreateVpc",
                "ec2:ModifyVpcAttribute",
                "ec2:CreateVpcEndpoint",
                # Add subnets
                "ec2:CreateSubnet",
                "ec2:ModifySubnetAttribute",
                # Add InternetGateway
                "ec2:CreateInternetGateway",
                "ec2:AttachInternetGateway",
                "ec2:DescribeInternetGateways",
                # Connect InternetGateway to Internet
                "ec2:CreateRouteTable",
                "ec2:AssociateRouteTable",
                "ec2:CreateRoute",
                "ec2:ReplaceRoute",
                # NAT Gateway Setup
                "ec2:CreateNatGateway",
                "ec2:DescribeNatGateways",
            ],
            "Resource": "*",
        },
        {
            "Sid": "CleanupEC2",
            "Effect": "Allow",
            "Action": [
                # Anyscale runs this on the first time a cloud is configured.
                "ec2:DeleteSecurityGroup",
                "ec2:RevokeSecurityGroupIngress",
                "ec2:RevokeSecurityGroupEgress",
                # Remove VPC
                "ec2:DeleteVpc",
                "ec2:DeleteVpcEndpoints",
                # Remove subnets
                "ec2:DeleteSubnet",
                # Remove InternetGateway
                "ec2:DeleteInternetGateway",
                "ec2:DetachInternetGateway",
                # Disconnect InternetGateway
                "ec2:DeleteRouteTable",
                "ec2:DisassociateRouteTable",
                "ec2:DeleteRoute",
                # Remove NATGateway
                "ec2:DeleteNatGateway",
            ],
            "Resource": "*",
        },
    ],
}

# Used for experimental readonly access to SSM
ANYSCALE_SSM_READONLY_ACCESS_POLICY_DOCUMENT = {
    "Version": "2012-10-17",
    "Statement": {
        "Sid": "SecretsManagerReadOnly",
        "Effect": "Allow",
        "Action": [
            "secretsmanager:GetSecretValue",
            "secretsmanager:DescribeSecret",
            "secretsmanager:ListSecrets",
        ],
        "Resource": "*",
    },
}

# Used for experimental read and write access to SSM
ANYSCALE_SSM_READ_WRITE_ACCESS_POLICY_DOCUMENT = {
    "Version": "2012-10-17",
    "Statement": {
        "Sid": "SecretsManagerReadWrite",
        "Effect": "Allow",
        "Action": [
            "secretsmanager:CreateSecret",
            "secretsmanager:PutSecretValue",
            "secretsmanager:GetSecretValue",
            "secretsmanager:DescribeSecret",
            "secretsmanager:ListSecrets",
        ],
        "Resource": "*",
    },
}


def get_anyscale_aws_iam_assume_role_policy(
    anyscale_aws_account: str,
) -> Dict[str, Any]:
    return {
        "Version": "2012-10-17",
        "Statement": [
            {
                "Sid": "AnyscaleControlPlaneAssumeRole",
                "Effect": "Allow",
                "Principal": {"AWS": f"arn:aws:iam::{anyscale_aws_account}:root"},
                "Action": "sts:AssumeRole",
            }
        ],
    }



import concurrent.futures
from contextlib import contextmanager
from datetime import datetime
from enum import Enum
import time
from typing import Any, Callable, Dict, List, Optional, Set

from click import ClickException
from rich.console import Group
from rich.live import Live
from rich.markup import escape
from rich.panel import Panel
from rich.progress import (
    BarColumn,
    Progress,
    TaskID,
    TextColumn,
    TimeElapsedColumn,
)
from rich.table import Table

from anyscale.cli_logger import LogsLogger
from anyscale.client.openapi_client.models import (
    ApplyProductionServiceV2Model,
    CloudAnalyticsEventName,
    CloudProviders,
    CreateExperimentalWorkspace,
    ServiceEventCurrentState,
    SessionState,
)
from anyscale.cluster_env import get_default_cluster_env_build
from anyscale.controllers.base_controller import BaseController
from anyscale.project import get_default_project
from anyscale.sdk.anyscale_client.models import (
    ComputeNodeType,
    ComputeTemplateQuery,
    CreateClusterCompute,
    CreateClusterComputeConfig,
    ServiceConfig,
)
from anyscale.util import confirm, get_endpoint
from anyscale.utils.cloud_utils import CloudEventProducer


POLL_INTERVAL_SECONDS = 10
WORKSPACE_VERIFICATION_TIMEOUT_MINUTES = 10
SERVICE_VERIFICATION_TIMEOUT_MINUTES = 15  # for a single rollout

# default values for cluster compute config
MAXIMUM_UPTIME_MINUTES = 15
IDLE_TERMINATION_MINUTES = 5
HEAD_NODE_TYPE_AWS = "m5.xlarge"  # on demand price ~$0.20 per hour
HEAD_NODE_TYPE_GCP = "n1-standard-2"  # on demand price ~$0.11 per hour


class CloudFunctionalVerificationType(str, Enum):
    WORKSPACE = "WORKSPACE"
    SERVICE = "SERVICE"


class CloudFunctionalVerificationTask:
    def __init__(self, task_id: TaskID, description: str):
        self.task_id = task_id
        self.description = description
        self.succeeded = False
        self.completed = False

    def update(self, succeeded: bool, description: str, completed: bool = False):
        self.description = description
        self.succeeded = succeeded
        self.completed = completed


class CloudFunctionalVerificationController(BaseController):
    def __init__(
        self,
        cloud_event_producer: CloudEventProducer,
        log: Optional[LogsLogger] = None,
        initialize_auth_api_client: bool = True,
    ):
        if log is None:
            log = LogsLogger()

        super().__init__(initialize_auth_api_client=initialize_auth_api_client)
        self.cloud_event_producer = cloud_event_producer
        self.log = log

        self.event_log_num: Dict[CloudFunctionalVerificationType, int] = {
            CloudFunctionalVerificationType.WORKSPACE: 0,
            CloudFunctionalVerificationType.SERVICE: 0,
        }

        # Used for rich live console
        self.step_progress: Dict[CloudFunctionalVerificationType, Progress] = {}
        self.overall_progress: Dict[CloudFunctionalVerificationType, Progress] = {}
        self.task_ids: Dict[CloudFunctionalVerificationType, TaskID] = {}
        self.event_log_tables: Dict[CloudFunctionalVerificationType, Table] = {}

    @staticmethod
    def get_head_node_type(cloud_provider: CloudProviders) -> str:
        """
        Get the default head node type for the given cloud provider.
        """
        if cloud_provider == CloudProviders.AWS:
            return HEAD_NODE_TYPE_AWS
        elif cloud_provider == CloudProviders.GCP:
            return HEAD_NODE_TYPE_GCP
        raise ClickException(f"Unsupported cloud provider: {cloud_provider}")

    def get_or_create_cluster_compute(
        self, cloud_id: str, cloud_provider: CloudProviders
    ) -> str:
        """
        Get or create a cluster compute for cloud functional verification
        """
        cluster_compute_name = f"functional_verification_{cloud_id}"

        cluster_computes = self.api_client.search_compute_templates_api_v2_compute_templates_search_post(
            ComputeTemplateQuery(
                orgwide=True,
                name={"equals": cluster_compute_name},
                include_anonymous=True,
            )
        ).results
        if len(cluster_computes) > 0:
            return cluster_computes[0].id

        head_node_instance_type = self.get_head_node_type(cloud_provider)
        # no cluster compute found, create one
        cluster_compute_config = CreateClusterComputeConfig(
            cloud_id=cloud_id,
            max_workers=0,
            allowed_azs=["any"],
            head_node_type=ComputeNodeType(
                name="head_node_type", instance_type=head_node_instance_type,
            ),
            maximum_uptime_minutes=MAXIMUM_UPTIME_MINUTES,
            idle_termination_minutes=IDLE_TERMINATION_MINUTES,
            worker_node_types=[],
        )
        if cloud_provider == CloudProviders.AWS:
            cluster_compute_config.aws_advanced_configurations_json = {
                "TagSpecifications": [
                    {
                        "ResourceType": "instance",
                        "Tags": [
                            {"Key": "cloud_functional_verification", "Value": cloud_id,}
                        ],
                    }
                ]
            }
        elif cloud_provider == CloudProviders.GCP:
            cluster_compute_config.gcp_advanced_configurations_json = {
                "instance_properties": {
                    "labels": {"cloud_functional_verification": cloud_id},
                }
            }

        cluster_compute = self.anyscale_api_client.create_cluster_compute(
            CreateClusterCompute(
                name=cluster_compute_name,
                config=cluster_compute_config,
                anonymous=True,
            )
        ).result
        return cluster_compute.id

    def _prepare_verification(self, cloud_id: str, cloud_provider: CloudProviders):
        """
        Generate the required parameters for cloud functional verification.
        """
        cluster_compute_id = self.get_or_create_cluster_compute(
            cloud_id, cloud_provider
        )

        cluster_env_build_id = get_default_cluster_env_build(
            self.api_client, self.anyscale_api_client
        ).id

        project_id = get_default_project(
            self.api_client, self.anyscale_api_client, parent_cloud_id=cloud_id
        ).id

        return cluster_compute_id, cluster_env_build_id, project_id

    @contextmanager
    def _create_task(self, function: CloudFunctionalVerificationType, description: str):
        """
        Create a task on the console for cloud functional verification
        """
        task_id = self.step_progress[function].add_task(description)
        task = CloudFunctionalVerificationTask(task_id, description)
        try:
            yield task
        finally:
            self._update_console(
                task.succeeded, function, task_id, task.description, task.completed
            )

    def create_workspace(self, cloud_id: str, cloud_provider: CloudProviders):
        """
        Create a workspace for cloud functional verification
        """
        (
            cluster_compute_id,
            cluster_env_build_id,
            project_id,
        ) = self._prepare_verification(cloud_id, cloud_provider)

        create_workspace_arg = CreateExperimentalWorkspace(
            name=f"fxnvrf_{cloud_id}_{datetime.utcnow().strftime('%Y%m%d%H%M%S')}",
            description=f"workspace for cloud {cloud_id} functional verification",
            project_id=project_id,
            cloud_id=cloud_id,
            compute_config_id=cluster_compute_id,
            cluster_environment_build_id=cluster_env_build_id,
            idle_timeout_minutes=IDLE_TERMINATION_MINUTES,
        )

        workspace = self.api_client.create_workspace_api_v2_experimental_workspaces_post(
            create_workspace_arg
        ).result

        return workspace

    def verify_workspace(self, cloud_id: str, cloud_provider: CloudProviders) -> bool:
        """
        Verifies that the workspace is setup correctly on the given cloud.
        """
        # Create workspace
        with self._create_task(
            CloudFunctionalVerificationType.WORKSPACE, "Creating workspace..."
        ) as create_workspace_task:
            try:
                workspace = self.create_workspace(cloud_id, cloud_provider)
            except ClickException as e:
                self.cloud_event_producer.produce(
                    CloudAnalyticsEventName.WORKSPACE_FUNCTIONAL_VERIFIED,
                    succeeded=False,
                    internal_error=str(e),
                )
                create_workspace_task.update(
                    False, f"[bold red]Failed to create workspace: {e}"
                )
                return False
            url = get_endpoint(f"/workspaces/{workspace.id}/{workspace.cluster_id}")
            create_workspace_task.update(
                True, f"[bold green]Workspace created at [link={url}]{url}[/link]"
            )

        # Wait until workspace is active
        def get_workspace_status(workspace_id):
            return self.api_client.get_workspace_api_v2_experimental_workspaces_workspace_id_get(
                workspace_id
            ).result.state

        allowed_status_set = {
            SessionState.RUNNING,
            SessionState.STARTINGUP,
            SessionState.AWAITINGSTARTUP,
        }

        with self._create_task(
            CloudFunctionalVerificationType.WORKSPACE,
            "Waiting for workspace to become active...",
        ) as wait_task:
            try:
                self.poll_until_active(
                    CloudFunctionalVerificationType.WORKSPACE,
                    workspace,
                    get_workspace_status,
                    SessionState.RUNNING,
                    allowed_status_set,
                    wait_task,
                    WORKSPACE_VERIFICATION_TIMEOUT_MINUTES,
                )
            except ClickException as e:
                self.cloud_event_producer.produce(
                    CloudAnalyticsEventName.WORKSPACE_FUNCTIONAL_VERIFIED,
                    succeeded=False,
                    internal_error=str(e),
                )
                wait_task.update(
                    False,
                    f"[bold red]Error: {e}. Please click on the URL above to check the logs.",
                )
                return False
            wait_task.update(True, "[bold green]Workspace is active.")

        # terminate workspace
        with self._create_task(
            CloudFunctionalVerificationType.WORKSPACE, "Terminating workspace..."
        ) as terminate_workspace_task:
            try:
                # terminate the cluster leads to workspace termination
                self.anyscale_api_client.terminate_cluster(workspace.cluster_id, {})
                allowed_status_set = {
                    SessionState.TERMINATED,
                    SessionState.TERMINATING,
                    SessionState.RUNNING,
                }
                self.poll_until_active(
                    CloudFunctionalVerificationType.WORKSPACE,
                    workspace,
                    get_workspace_status,
                    SessionState.TERMINATED,
                    allowed_status_set,
                    terminate_workspace_task,
                    WORKSPACE_VERIFICATION_TIMEOUT_MINUTES,
                )
                # archive workspace
                self.api_client.delete_workspace_api_v2_experimental_workspaces_workspace_id_delete(
                    workspace.id
                )
            except ClickException as e:
                terminate_workspace_task.update(
                    False,
                    f"[bold red]Failed to terminate workspace: {e}",
                    completed=True,
                )
                self.cloud_event_producer.produce(
                    CloudAnalyticsEventName.WORKSPACE_FUNCTIONAL_VERIFIED,
                    succeeded=False,
                    internal_error=str(e),
                )
                return False
            terminate_workspace_task.update(
                True, "[bold green]Workspace terminated.", completed=True
            )

        self.cloud_event_producer.produce(
            CloudAnalyticsEventName.WORKSPACE_FUNCTIONAL_VERIFIED, succeeded=True,
        )
        return True

    def rollout_service(
        self,
        cloud_id: str,
        cloud_provider: CloudProviders,
        *,
        version: str = "v1",
        service_name: str = None,
    ):
        """
        Roll out a service for cloud functional verification
        """
        (
            cluster_compute_id,
            cluster_env_build_id,
            project_id,
        ) = self._prepare_verification(cloud_id, cloud_provider)

        service_name = (
            service_name
            or f"fxnvrf_{cloud_id}_{datetime.utcnow().strftime('%Y%m%d%H%M%S')}"
        )

        service_config = ApplyProductionServiceV2Model(
            name=service_name,
            description=f"service for cloud {cloud_id} functional verification",
            project_id=project_id,
            ray_serve_config={
                "import_path": "serve_hello:entrypoint",
                "runtime_env": {
                    "working_dir": "https://github.com/anyscale/docs_examples/archive/refs/heads/main.zip",
                    "env_vars": {
                        "SERVE_RESPONSE_MESSAGE": f"cloud functional verification {version}",
                    },
                },
            },
            version=version,
            build_id=cluster_env_build_id,
            compute_config_id=cluster_compute_id,
            config=ServiceConfig(
                max_uptime_timeout_sec=SERVICE_VERIFICATION_TIMEOUT_MINUTES * 60,
            ),
        )

        # Rollout service
        service = self.api_client.apply_service_v2_api_v2_services_v2_apply_put(
            service_config
        ).result

        return service

    def verify_service(self, cloud_id: str, cloud_provider: CloudProviders) -> bool:
        """
        Verifies that the service can be deployed and updated on the given cloud.
        """
        # Deploy service
        with self._create_task(
            CloudFunctionalVerificationType.SERVICE, "Deploying service..."
        ) as deploy_service_task:
            try:
                service = self.rollout_service(cloud_id, cloud_provider, version="v1")
            except ClickException as e:
                self.cloud_event_producer.produce(
                    CloudAnalyticsEventName.SERVICE_FUNCTIONAL_VERIFIED,
                    succeeded=False,
                    internal_error=str(e),
                )
                deploy_service_task.update(
                    False, f"[bold red]Failed to deploy service: {e}"
                )
                return False
            url = get_endpoint(f"/services/{service.id}")
            deploy_service_task.update(
                True, f"[bold green]Service deployed at [link={url}]{url}[/link]"
            )

        # Wait until service is active
        def get_service_status(service_id):
            return self.api_client.get_service_api_v2_services_v2_service_id_get(
                service_id
            ).result.current_state

        allowed_status_set = {
            ServiceEventCurrentState.RUNNING,
            ServiceEventCurrentState.UPDATING,
            ServiceEventCurrentState.STARTING,
        }

        with self._create_task(
            CloudFunctionalVerificationType.SERVICE,
            "Waiting for service to become active...",
        ) as wait_task:
            try:
                self.poll_until_active(
                    CloudFunctionalVerificationType.SERVICE,
                    service,
                    get_service_status,
                    ServiceEventCurrentState.RUNNING,
                    allowed_status_set,
                    wait_task,
                    SERVICE_VERIFICATION_TIMEOUT_MINUTES,
                )
            except ClickException as e:
                self.cloud_event_producer.produce(
                    CloudAnalyticsEventName.SERVICE_FUNCTIONAL_VERIFIED,
                    succeeded=False,
                    internal_error=str(e),
                )
                wait_task.update(
                    False,
                    f"[bold red]Error: {e}. Please click on the URL above to check the logs.",
                )
                return False
            wait_task.update(True, "[bold green]Service is active.")

        # Update service
        with self._create_task(
            CloudFunctionalVerificationType.SERVICE, "Updating service..."
        ) as update_service_task:
            try:
                service = self.rollout_service(
                    cloud_id, cloud_provider, version="v2", service_name=service.name
                )
            except ClickException as e:
                self.cloud_event_producer.produce(
                    CloudAnalyticsEventName.SERVICE_FUNCTIONAL_VERIFIED,
                    succeeded=False,
                    internal_error=str(e),
                )
                update_service_task.update(
                    False, f"[bold red]Failed to update service: {e}"
                )
                return False

            # Wait until service is active
            allowed_status_set.add(ServiceEventCurrentState.ROLLING_OUT)
            try:
                self.poll_until_active(
                    CloudFunctionalVerificationType.SERVICE,
                    service,
                    get_service_status,
                    ServiceEventCurrentState.RUNNING,
                    allowed_status_set,
                    update_service_task,
                    SERVICE_VERIFICATION_TIMEOUT_MINUTES,
                )
            except ClickException as e:
                self.cloud_event_producer.produce(
                    CloudAnalyticsEventName.SERVICE_FUNCTIONAL_VERIFIED,
                    succeeded=False,
                    internal_error=str(e),
                )
                update_service_task.update(
                    False,
                    f"[bold red]Error: {e}. Please click on the URL above to check the logs.",
                )
                return False
            update_service_task.update(
                True, "[bold green]Service upgraded successfully."
            )

        # Terminate service
        with self._create_task(
            CloudFunctionalVerificationType.SERVICE, "Terminating service..."
        ) as terminate_service_task:
            try:
                self.api_client.terminate_service_api_v2_services_v2_service_id_terminate_post(
                    service.id
                )
                allowed_status_set = {
                    ServiceEventCurrentState.TERMINATED,
                    ServiceEventCurrentState.TERMINATING,
                    ServiceEventCurrentState.RUNNING,
                }
                self.poll_until_active(
                    CloudFunctionalVerificationType.SERVICE,
                    service,
                    get_service_status,
                    ServiceEventCurrentState.TERMINATED,
                    allowed_status_set,
                    terminate_service_task,
                    SERVICE_VERIFICATION_TIMEOUT_MINUTES,
                )
            except ClickException as e:
                terminate_service_task.update(
                    False,
                    f"[bold red]Failed to terminate service: {e}",
                    completed=True,
                )
                self.cloud_event_producer.produce(
                    CloudAnalyticsEventName.SERVICE_FUNCTIONAL_VERIFIED,
                    succeeded=False,
                    internal_error=str(e),
                )
                return False
            terminate_service_task.update(
                True, "[bold green]Service terminated.", completed=True
            )

        self.cloud_event_producer.produce(
            CloudAnalyticsEventName.SERVICE_FUNCTIONAL_VERIFIED, succeeded=True,
        )
        return True

    def poll_until_active(  # noqa: PLR0913
        self,
        function_type: CloudFunctionalVerificationType,
        function: Any,
        get_current_status: Callable[[str], Any],
        goal_status: Any,
        allowed_status_set: Set[Any],
        wait_task: CloudFunctionalVerificationTask,
        timeout_minutes: int,
    ) -> bool:
        """
        Polling until it is active.
        """
        start_time = time.time()
        end_time = start_time + timeout_minutes * 60
        while time.time() < end_time:
            time.sleep(POLL_INTERVAL_SECONDS)
            try:
                current_status = get_current_status(function.id)
                self._render_event_log(function_type, function)
            except ClickException as e:
                raise ClickException(
                    f"Failed to get {function_type.lower()} status: {e}"
                ) from None
            self._update_task_in_step_progress(
                function_type,
                wait_task.task_id,
                f"{wait_task.description} [{time.strftime('%H:%M:%S', time.localtime())}] Current status: {current_status}",
            )
            if current_status == goal_status:
                return True
            if current_status not in allowed_status_set:
                raise ClickException(
                    f"{function_type.capitalize()} is in an unexpected state: {current_status}"
                )
        raise ClickException(
            f"Timed out waiting for {function_type.lower()} to become active"
        )

    def verify(
        self,
        function: CloudFunctionalVerificationType,
        cloud_id: str,
        cloud_provider: CloudProviders,
    ) -> bool:
        """
        Kick off a single functional verification given the function type.
        """
        if function == CloudFunctionalVerificationType.WORKSPACE:
            return self.verify_workspace(cloud_id, cloud_provider)
        elif function == CloudFunctionalVerificationType.SERVICE:
            return self.verify_service(cloud_id, cloud_provider)
        return False

    def _update_console(
        self,
        succeeded: bool,
        function: CloudFunctionalVerificationType,
        task_id: TaskID,
        description: str,
        completed: bool = False,
    ):
        """
        Update the console based on the verification result
        """
        self._update_overall_progress(succeeded, function, completed)
        self._finish_task_in_step_progress(function, task_id, description)

    def _update_task_in_step_progress(
        self,
        function: CloudFunctionalVerificationType,
        task_id: TaskID,
        description: str,
    ) -> None:
        """
        Update the task description in step progress
        """
        self.step_progress[function].update(task_id, description=description)

    def _finish_task_in_step_progress(
        self,
        function: CloudFunctionalVerificationType,
        task_id: TaskID,
        description: str,
    ) -> None:
        """
        Finish a task in step progress and update the description
        """
        self.step_progress[function].stop_task(task_id)
        self.step_progress[function].update(task_id, description=description)

    def _update_overall_progress(
        self,
        verification_result: bool,
        function: CloudFunctionalVerificationType,
        completed: bool = False,
    ) -> None:
        """
        Update overall progress based on the verification result
        """
        if verification_result:
            self.overall_progress[function].advance(self.task_ids[function], 1)
            if completed:
                self.overall_progress[function].update(
                    self.task_ids[function],
                    description=f"[bold green]{function.capitalize()} verification succeeded!",
                )
        else:
            self.overall_progress[function].stop_task(self.task_ids[function])
            self.overall_progress[function].update(
                self.task_ids[function],
                description=f"[bold red]{function.capitalize()} verification failed. Please check the logs and terminate the {function.lower()} manually.",
            )

    def _format_service_event_log(self, event_log: Any) -> str:
        description = event_log.message.replace("\n", ". ") if event_log.message else ""
        return f"[bold cyan]{event_log.created_at}[/bold cyan] [bold blue]{event_log.event_type}[/bold blue] [bold red]{description}[/bold red]"

    def _render_event_log(
        self, function_type: CloudFunctionalVerificationType, function: Any
    ):
        if function_type == CloudFunctionalVerificationType.WORKSPACE:
            workspace_log = self.api_client.get_startup_logs_api_v2_sessions_session_id_startup_logs_get(
                function.cluster_id, self.event_log_num[function_type], 100000000
            ).result
            if workspace_log.num_lines == self.event_log_num[function_type]:
                # no new logs
                return
            self.event_log_tables[function_type].add_row(escape(workspace_log.lines))
            self.event_log_num[function_type] = workspace_log.num_lines
        elif function_type == CloudFunctionalVerificationType.SERVICE:
            service_events = self.api_client.get_service_events_api_v2_services_v2_service_id_events_get(
                function.id
            ).results
            starting_pos = len(service_events) - self.event_log_num[function_type] - 1
            for idx in range(starting_pos, -1, -1):
                self.event_log_tables[function_type].add_row(
                    self._format_service_event_log(service_events[idx])
                )
            self.event_log_num[function_type] = len(service_events)
        else:
            # we should never enter this branch
            return

    def get_live_console(
        self, functions_to_verify: List[CloudFunctionalVerificationType]
    ) -> Live:
        """
        Get a live console for cloud functional verification.

        Each functional verification contains 3 components:
        1. step progress panel: shows the progress bars for each functional verification step
        2. event log table: event logs for each functional verification
        3. overall progress: overall progress bar for the functional verification
        """
        progress_group = []
        steps = {
            CloudFunctionalVerificationType.WORKSPACE: 3,
            CloudFunctionalVerificationType.SERVICE: 4,
        }
        for function in functions_to_verify:
            progress_table = Table.grid(expand=True)
            step_progress = Progress(
                TimeElapsedColumn(), TextColumn("{task.description}"),
            )
            self.step_progress[function] = step_progress
            event_log_table = Table(box=None)
            self.event_log_tables[function] = event_log_table
            progress_table.add_row(
                Panel(step_progress, title=f"{function.lower()} verification")
            )
            progress_table.add_row(
                Panel(event_log_table, title=f"{function.lower()} event logs")
            )

            progress_group.append(progress_table)

            overall_progress = Progress(
                TimeElapsedColumn(), BarColumn(), TextColumn("{task.description}")
            )
            self.overall_progress[function] = overall_progress
            progress_group.append(overall_progress)
            task_id = overall_progress.add_task("", total=steps[function])
            overall_progress.update(
                task_id, description=f"[bold #AAAAAA]Verifying {function.lower()}..."
            )
            self.task_ids[function] = task_id
        return Live(Group(*progress_group))

    def start_verification(
        self,
        cloud_id: str,
        cloud_provider: CloudProviders,
        functions_to_verify: List[CloudFunctionalVerificationType],
        yes: bool = False,
    ) -> bool:
        """
        Starts cloud functional verification
        """
        with self.log.spinner("Starting functional verification..."):
            self._prepare_verification(cloud_id, cloud_provider)

        self.log.info(
            f"Functional verification for {', '.join(functions_to_verify)} is about to begin."
        )

        confirmation_message = [
            f"It will spin up one {self.get_head_node_type(cloud_provider)} instance for each function.",
        ]

        if CloudFunctionalVerificationType.WORKSPACE in functions_to_verify:
            confirmation_message.append(
                "Workspace verification takes about 5 minutes. "
                "The approximate cost from the cloud provider is about $0.02 (it varies from cloud provider and region)."
            )

        if CloudFunctionalVerificationType.SERVICE in functions_to_verify:
            service_time_estimation = {
                CloudProviders.AWS: 8,
                CloudProviders.GCP: 20,
            }
            confirmation_message.append(
                f"Service verification takes about {service_time_estimation[cloud_provider]} minutes. "
                "The approximate cost from the cloud provider is about $0.10 (it varies from cloud provider and region)."
            )

        confirmation_message.append(
            "The instances will be terminated after verification. Do you want to continue?"
        )

        confirm(
            "\n".join(confirmation_message), yes,
        )

        verification_results: List[bool] = []
        with self.get_live_console(
            functions_to_verify
        ), concurrent.futures.ThreadPoolExecutor(
            max_workers=len(CloudFunctionalVerificationType)
        ) as executor:
            futures = {
                executor.submit(
                    self.verify, function, cloud_id, cloud_provider
                ): function
                for function in functions_to_verify
            }

            # Wait for all verifications to complete
            for future in concurrent.futures.as_completed(futures):
                function = futures[future]
                try:
                    verification_result = future.result()
                    verification_results.append(verification_result)
                except Exception as e:  # noqa: BLE001
                    self._update_console(
                        False,
                        function,
                        self.task_ids[function],
                        f"[bold red]Failed to verify {function.lower()}: {e}",
                        completed=True,
                    )
                    verification_results.append(False)
        return all(verification_results)



# coding: utf-8

"""
    Managed Ray API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from openapi_client.configuration import Configuration


class LogsOutput(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'lines': 'str',
        'num_lines': 'int',
        'start_line': 'int',
        'finished': 'bool'
    }

    attribute_map = {
        'lines': 'lines',
        'num_lines': 'num_lines',
        'start_line': 'start_line',
        'finished': 'finished'
    }

    def __init__(self, lines=None, num_lines=None, start_line=None, finished=None, local_vars_configuration=None):  # noqa: E501
        """LogsOutput - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._lines = None
        self._num_lines = None
        self._start_line = None
        self._finished = None
        self.discriminator = None

        self.lines = lines
        self.num_lines = num_lines
        self.start_line = start_line
        self.finished = finished

    @property
    def lines(self):
        """Gets the lines of this LogsOutput.  # noqa: E501


        :return: The lines of this LogsOutput.  # noqa: E501
        :rtype: str
        """
        return self._lines

    @lines.setter
    def lines(self, lines):
        """Sets the lines of this LogsOutput.


        :param lines: The lines of this LogsOutput.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and lines is None:  # noqa: E501
            raise ValueError("Invalid value for `lines`, must not be `None`")  # noqa: E501

        self._lines = lines

    @property
    def num_lines(self):
        """Gets the num_lines of this LogsOutput.  # noqa: E501


        :return: The num_lines of this LogsOutput.  # noqa: E501
        :rtype: int
        """
        return self._num_lines

    @num_lines.setter
    def num_lines(self, num_lines):
        """Sets the num_lines of this LogsOutput.


        :param num_lines: The num_lines of this LogsOutput.  # noqa: E501
        :type: int
        """
        if self.local_vars_configuration.client_side_validation and num_lines is None:  # noqa: E501
            raise ValueError("Invalid value for `num_lines`, must not be `None`")  # noqa: E501

        self._num_lines = num_lines

    @property
    def start_line(self):
        """Gets the start_line of this LogsOutput.  # noqa: E501


        :return: The start_line of this LogsOutput.  # noqa: E501
        :rtype: int
        """
        return self._start_line

    @start_line.setter
    def start_line(self, start_line):
        """Sets the start_line of this LogsOutput.


        :param start_line: The start_line of this LogsOutput.  # noqa: E501
        :type: int
        """
        if self.local_vars_configuration.client_side_validation and start_line is None:  # noqa: E501
            raise ValueError("Invalid value for `start_line`, must not be `None`")  # noqa: E501

        self._start_line = start_line

    @property
    def finished(self):
        """Gets the finished of this LogsOutput.  # noqa: E501


        :return: The finished of this LogsOutput.  # noqa: E501
        :rtype: bool
        """
        return self._finished

    @finished.setter
    def finished(self, finished):
        """Sets the finished of this LogsOutput.


        :param finished: The finished of this LogsOutput.  # noqa: E501
        :type: bool
        """
        if self.local_vars_configuration.client_side_validation and finished is None:  # noqa: E501
            raise ValueError("Invalid value for `finished`, must not be `None`")  # noqa: E501

        self._finished = finished

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, LogsOutput):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, LogsOutput):
            return True

        return self.to_dict() != other.to_dict()



# coding: utf-8

"""
    Managed Ray API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from openapi_client.configuration import Configuration


class LogDownloadConfig(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'next_page_token': 'str',
        'page_size': 'int',
        'ttl_seconds': 'int',
        'use_bearer_token': 'bool'
    }

    attribute_map = {
        'next_page_token': 'next_page_token',
        'page_size': 'page_size',
        'ttl_seconds': 'ttl_seconds',
        'use_bearer_token': 'use_bearer_token'
    }

    def __init__(self, next_page_token=None, page_size=1000, ttl_seconds=14400, use_bearer_token=False, local_vars_configuration=None):  # noqa: E501
        """LogDownloadConfig - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._next_page_token = None
        self._page_size = None
        self._ttl_seconds = None
        self._use_bearer_token = None
        self.discriminator = None

        if next_page_token is not None:
            self.next_page_token = next_page_token
        if page_size is not None:
            self.page_size = page_size
        if ttl_seconds is not None:
            self.ttl_seconds = ttl_seconds
        if use_bearer_token is not None:
            self.use_bearer_token = use_bearer_token

    @property
    def next_page_token(self):
        """Gets the next_page_token of this LogDownloadConfig.  # noqa: E501

        The next page token (requests to the Log Download API are paginated).  # noqa: E501

        :return: The next_page_token of this LogDownloadConfig.  # noqa: E501
        :rtype: str
        """
        return self._next_page_token

    @next_page_token.setter
    def next_page_token(self, next_page_token):
        """Sets the next_page_token of this LogDownloadConfig.

        The next page token (requests to the Log Download API are paginated).  # noqa: E501

        :param next_page_token: The next_page_token of this LogDownloadConfig.  # noqa: E501
        :type: str
        """

        self._next_page_token = next_page_token

    @property
    def page_size(self):
        """Gets the page_size of this LogDownloadConfig.  # noqa: E501

        Page size (default: 1000 items/page).  # noqa: E501

        :return: The page_size of this LogDownloadConfig.  # noqa: E501
        :rtype: int
        """
        return self._page_size

    @page_size.setter
    def page_size(self, page_size):
        """Sets the page_size of this LogDownloadConfig.

        Page size (default: 1000 items/page).  # noqa: E501

        :param page_size: The page_size of this LogDownloadConfig.  # noqa: E501
        :type: int
        """

        self._page_size = page_size

    @property
    def ttl_seconds(self):
        """Gets the ttl_seconds of this LogDownloadConfig.  # noqa: E501

        The time-to-live of the presigned url in seconds (default: 4 hours).  # noqa: E501

        :return: The ttl_seconds of this LogDownloadConfig.  # noqa: E501
        :rtype: int
        """
        return self._ttl_seconds

    @ttl_seconds.setter
    def ttl_seconds(self, ttl_seconds):
        """Sets the ttl_seconds of this LogDownloadConfig.

        The time-to-live of the presigned url in seconds (default: 4 hours).  # noqa: E501

        :param ttl_seconds: The ttl_seconds of this LogDownloadConfig.  # noqa: E501
        :type: int
        """

        self._ttl_seconds = ttl_seconds

    @property
    def use_bearer_token(self):
        """Gets the use_bearer_token of this LogDownloadConfig.  # noqa: E501

        Using access token allows us to avoid presigning urls. This is much faster in most cases. This is only supported in GCE. Caller must use the access_token returned from the API to download the logs. Set the access_token as a Bearer token in the Authorization header.  # noqa: E501

        :return: The use_bearer_token of this LogDownloadConfig.  # noqa: E501
        :rtype: bool
        """
        return self._use_bearer_token

    @use_bearer_token.setter
    def use_bearer_token(self, use_bearer_token):
        """Sets the use_bearer_token of this LogDownloadConfig.

        Using access token allows us to avoid presigning urls. This is much faster in most cases. This is only supported in GCE. Caller must use the access_token returned from the API to download the logs. Set the access_token as a Bearer token in the Authorization header.  # noqa: E501

        :param use_bearer_token: The use_bearer_token of this LogDownloadConfig.  # noqa: E501
        :type: bool
        """

        self._use_bearer_token = use_bearer_token

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, LogDownloadConfig):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, LogDownloadConfig):
            return True

        return self.to_dict() != other.to_dict()
